import java.net.*; 
import java.io.*; 
import java.util.*;

/** 
*  UsaProxy - HTTP proxy for tracking and logging of user interactions on websites
   Copyright (C) 2006  Monika Wnuk - Media Informatics Group at the University of Munich

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

*
* Class ClientRequest handles a single client request
* Request reading, forwarding to the web server/remote proxy
* and response processing are managed by this class.
* In Addition, log requests and file forwarding requests are
* routed to the instances of the respective classes,
* namely LogFile and ProxyScript
*/
public class ClientRequest extends Thread {

	private InetAddress 			remoteProxy;					  /** remote proxy data */
    private int 					remotePort;						  /** remote proxy data */
	private Socket 				clientConnect, serverConnect; 
    private InputStream 			clientIn,serverIn; 
    private OutputStream 			clientOut,serverOut;
    private String 				get,host;
    private Hashtable				requestHeaders, responseHeaders;  /**  request and response headers
    																	*  are easy to modify being stored
    																	*  in a hashtable */
    private boolean 				isHtmlStream, isNotFoundMessage;
    private String					getKey;						  	  /**  contains requested file path */
    private String					statusKey;						  /**  contains status line of response message */
    private StringBuffer 			data;
    
    private LogFile 				logFile;						  /**  contains synchronized log()-method */
    private ProxyScript			proxyScript;					  /**  contains synchronized sendFile()-method */
    private boolean 				proxyScriptRequested;
    private boolean 				logRequested;
    private String 				logData;						  /**  contains the data to log retrieved
    																	*  from the log request argument */
    private HTTPTraffic			httpTraffic;					  /**  provides access to index and txt-file for storing
    																	*  request and response data to */
    private int					httpTrafficIndex;				  /**  index for txt-file suffix and as identifier
    																	*  in httpTraffic log entry */
    final static boolean 			debug = false;
    
    
    
    /** Constructor: create ClientRequest with a given client socket*/
    public ClientRequest (Socket socket, LogFile logFile, ProxyScript proxyScript, HTTPTraffic httpTraffic) {

		this.clientConnect  	= socket;
		this.remoteProxy 		= null;
		this.remotePort 		= 0;
		this.logFile			= logFile;
		this.proxyScript		= proxyScript;
		this.httpTraffic		= httpTraffic;
		this.proxyScriptRequested 	= false;
		this.logRequested 			= false;
	
		/** start thread */
		start();
    }
    
    /** Constructor: create ClientRequest with a given client socket and remote proxy data */
    public ClientRequest (Socket socket, LogFile logFile, ProxyScript proxyScript, HTTPTraffic httpTraffic, InetAddress remoteProxy, int remotePort) {

		this.clientConnect  	= socket;
		this.remoteProxy 		= remoteProxy;
		this.remotePort 		= remotePort;
		this.logFile			= logFile;
		this.proxyScript		= proxyScript;
		this.httpTraffic		= httpTraffic;
		this.proxyScriptRequested 	= false;
		this.logRequested 			= false;
	
		/** start thread */
		start();
    }
    
 	/** run()-method which is called by start() */
    public void run () { 

		try { 
		    
		    /** Create streams */
		    clientIn  = clientConnect.getInputStream();
		    clientOut = clientConnect.getOutputStream();
		    
		    /** parse request and detect parameters */
		    get  	= "";
		    host 	= "";
		    requestHeaders 	= new Hashtable();
		    clientIn 		= processRequest(clientIn);
	
			/** if proxyscript.js is requested */
		    if (proxyScriptRequested) {
				proxyScript.send (clientOut, "proxyscript.js", clientConnect);
			}
		    /** if data is to be logged */
			else if (logRequested) {
				logFile.log(clientOut, logData, clientConnect);
			}
			/** all web server requests */
			else {
			    /** Open connection to the server */
			    try {	
			    	/** If remote proxy is used due to a network a socket to it is created*/
			    	if (remoteProxy != null) {
						serverConnect = new Socket(remoteProxy.getHostName(),remotePort);
			    	}
			    	/** open a direct connection to the webserver */
			    	else {
			    		URL url 	= new URL(get);
						int port 	= 80;
				
						/** Could port be detected ? */
						if (url.getPort() > 0) {
						    port = url.getPort();
						}
						
						serverConnect = new Socket(url.getHost(),port);
			    	}
			    }
			    /** if server could not be detected or didn't respond */
			    catch (Exception err) {
		
					System.err.println("\nAn ERROR occured while connecting to the server:\n"
									+ err );
		
					/** Send error message to client */
					PrintWriter out = new PrintWriter(new OutputStreamWriter(clientOut));
			
					out.println("HTTP/1.0 200 ");            /** version and status code */
					out.println("Content-Type: text/plain"); /** data will be sent as text */
					out.println();                           /** blank line */
			
					out.println("GET/POST:  " + get);
					out.println("HOST: " + host);
					out.println("Error!");
					out.println(err);
					out.flush();
			
					/** Close streams and sockets */
					if (clientOut     != null) {clientOut.close();}
					if (clientIn      != null) {clientIn.close();}
					if (clientConnect != null) {clientConnect.close ();}
					if (serverConnect != null) {serverConnect.close();}
			
					/** End of run */
					return;
					
			    }
			    /** Create server streams */
			    serverIn  = serverConnect.getInputStream();
			    serverOut = serverConnect.getOutputStream();
			    
			    if(httpTraffic.isLogHttpTraffic())
			    	httpTrafficIndex = httpTraffic.getHttpTrafficIndex(true);
			    
			    /** Create new Thread that manages the posting
			     *  of the client request to the web server */
			    ServerRequest serverRequest;
			    if (clientIn!=null)
			    	serverRequest = new ServerRequest(serverOut, requestHeaders, clientIn, this);
			    else
			    	serverRequest = new ServerRequest(serverOut, requestHeaders, this);
			    
			    /** Receive server response and send it to client */
				processResponse (serverIn);
			}  
		}
		catch (Exception e) {
		    System.err.println("\nAn ERROR occured while negotiating with the server:\n"
		    				+ e);
		    e.printStackTrace();
		    
		}
		finally { 
		    
		    /** Close streams and sockets */
		    try { 
			if (clientOut     != null) {clientOut.close();}
			if (clientIn      != null) {clientIn.close();}
			if (serverOut     != null) {serverOut.close();}
			if (serverIn      != null) {serverIn.close();}
			if (clientConnect != null) {clientConnect.close();}
			if (serverConnect != null) {serverConnect.close();}
		    }
		    catch (Exception ex) { /**...*/ }
		}
    }
    
    /** retrieve request headers and store in Hashtable requestHeaders
     *  return remaining input stream for further processing
     */
    private InputStream processRequest(InputStream in) {
    	
    	byte b 			= 0;
        int len 		= 0;
        int colonIndex	= -1;
        String name, value, tmpString;
        getKey = "";
        int returnCode	= 0;
        String method	= "";
        
        try {
	        if (in == null)
				in = new BufferedInputStream(clientConnect.getInputStream());
	
	        if (requestHeaders == null) 
	        	requestHeaders = new Hashtable();
	            
	        
	        /** Logic to read HTTP request headers */
	        boolean readTooMuch = false;
	//		System.out.println("");
	
	        for (ByteArrayOutputStream buf = new ByteArrayOutputStream(4097); ;) {
	            if (!readTooMuch) {
	                b = (byte) in.read();
	 //               System.out.print((char) b);
	            }
	            if (b == -1) {
	                break;
	            }
	            readTooMuch = false;
	            if ((b != '\r') && (b != '\n')) {
	                if ((b == ':') && (colonIndex == -1)) {
	                    colonIndex = len;
	                }
	                len++;
	                buf.write(b);
	            } else if (b == '\r') {
	                continue;
	            } else {    // b== '\n'
	                if (len == 0) {
	                    break;
	                }
	                b = (byte) in.read();
	  //              System.out.print((char) b);
	                readTooMuch = true;
	
	                /** A space or tab at the beginning of a line means the header continues */
	                if ((b == ' ') || (b == '\t')) {
	                    continue;
	                }
	                buf.close();
	                byte[] hdata = buf.toByteArray();
	                buf.reset();
	                if (colonIndex != -1) {
	                	tmpString = new String(hdata, 0, len, HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING);
	                	if (tmpString.toLowerCase().startsWith("get") || tmpString.toLowerCase().startsWith("post") || tmpString.toLowerCase().startsWith("connect ")) {
	                		name = tmpString;
	                		value = "";
	                		getKey = name;
	                	} else {
	                		name 	= new String(hdata, 0, colonIndex, HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING);
	                		value 	= new String(hdata, colonIndex + 2, len - 2 - colonIndex,
	                                    HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING);
	                	}
	                    colonIndex = -1;
	                } else {
	                    name = new String(hdata, 0, len,
	                                    HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING);
	                    value = "";
	                    
	                    if (name.toLowerCase().startsWith("get") || name.toLowerCase().startsWith("post") || name.toLowerCase().startsWith("connect "))
	                    	getKey = name;
	                }
	                String[ ] headerField = {name, value};
	                if(name.equals(getKey)) 	requestHeaders.put(name, headerField);
	                else						requestHeaders.put(name.toLowerCase(), headerField);
	                len = 0;
	            }
	        }
	        
	        
	        /** retrieve url */
			String url 		= getKey;
			requestHeaders.remove(getKey);
			if (url.toLowerCase().startsWith("get")) 
				method 	= "GET";
			else if (url.toLowerCase().startsWith("post"))	/** post */
				method 	= "POST";
			else  /** connect */
				method 	= "CONNECT";
			url = url.substring(method.length()).trim();
			if (method.equals("CONNECT")) method = "GET";
			
			/** delete HTTP version */
			if (url.toUpperCase().lastIndexOf("HTTP") >= 0) 	/** if HTTP version exists */
			    url = url.substring(0,url.toUpperCase().lastIndexOf("HTTP")).trim();
			else
			    url = url.trim();
			get = url;
			
			/** if js-file is ordered */
			if (url.toLowerCase().matches("http://lo.lo/proxyscript.js")) {
				proxyScriptRequested = true;
				return null;
			}
			
			/** if log request */
			if (url.toLowerCase().startsWith("http://lo.lo/img.jpg")) {
				logRequested = true;
				logData = url.substring(url.indexOf("?") + 1);
				return null;
			}
	
			/** Create new GET or POST-line:
			 ** GET http://host.xyz:80/index.html HTTP 1.0
			 ** cut off the host part of the URL and leave the filepath and name */
	/**		if(!(url.toLowerCase().startsWith("http")))
				url = "http://" + url;  */
			URL geturl = new URL(url);
	
			if (remoteProxy != null) {
				/** add new line with default HTTP/1.0 */
				requestHeaders.put(method + " " + geturl.toString() + " HTTP/1.0","");
				getKey = method + " " + geturl.toString() + " HTTP/1.0";
			}
			else {
				/** add new line with default HTTP/1.0 */
				requestHeaders.put(method + " " + geturl.getFile() + " HTTP/1.0","");
				getKey = method + " " + geturl.toString() + " HTTP/1.0";
			}
	
			/** Display message: Client and requested URL */
			if (debug) System.out.println("New REQUEST:");
			if (debug) System.out.println("From client " + clientConnect.getInetAddress().getHostName() + " (" + clientConnect.getInetAddress().getHostAddress() + ") with request GET " + url);
	        
	        /** retrieve host */
	        String[] headerFieldHost = (String[])requestHeaders.get(HTTPConstants.HEADER_HOST.toLowerCase());
		    host = headerFieldHost[1];
	        
	        /** add accept-encoding header value "identity" to avoid gzip content-encoding */
	        String[] headerFieldGZIP = {HTTPConstants.HEADER_ACCEPT_ENCODING,"identity"};
	        requestHeaders.put(HTTPConstants.HEADER_ACCEPT_ENCODING.toLowerCase(), headerFieldGZIP);
	        
	        /** add connection header value "close" */
	        String[] headerFieldConn = {HTTPConstants.HEADER_CONNECTION,"close"};
	        requestHeaders.put(HTTPConstants.HEADER_CONNECTION.toLowerCase(), headerFieldConn);
	        
	        /** remove proxy-connection and keep-alive header */
	        if(requestHeaders.containsKey("proxy-connection"))
	        	requestHeaders.remove("proxy-connection");
	        if(requestHeaders.containsKey("keep-alive"))
	        	requestHeaders.remove("keep-alive"); 
	        
	        /** add pragma header value "no-cache" to avoid caching */
	        String[] headerFieldPragma = {HTTPConstants.HEADER_PRAGMA,"no-cache"};
	        requestHeaders.put(HTTPConstants.HEADER_PRAGMA.toLowerCase(), headerFieldPragma);
	        
	        /** add UsaProxy header for Apache mod_rewrite */
	        String[] headerFieldUsaP = {HTTPConstants.HEADER_X_USAPROXY,"client"};
	        requestHeaders.put(HTTPConstants.HEADER_X_USAPROXY.toLowerCase(), headerFieldUsaP);
        
        } catch (IOException e) {
        	if (clientConnect  != null) {
        		System.err.println("\nAn ERROR occured while processing request: "
        					+ getKey + " from client " + clientConnect.getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while processing request: "
        					+ "\n"
							+ e );
        	}
		}
		if (method.equals("POST"))
			return in;
		else
			return null;
        
    }
    
    /** Read and parse response and retrieve parameters*/
	private void processResponse (InputStream in) {
    	
    	/** Read the header from the server response */
        responseHeaders 	= new Hashtable();
        
        in 					= readResponseHeaders(in);
        
        /** Check if html is being transmitted */
        String[] headerFieldType = (String[])responseHeaders.get(HTTPConstants.HEADER_CONTENT_TYPE.toLowerCase());
        String contentType = headerFieldType[1];
        if ((contentType != null) && (contentType.indexOf("text/html") != -1))
        	isHtmlStream = true;
        else isHtmlStream = false;
        
        /** Check if 404 message is being transmitted
         *  if true no enhancing necessary*/
        if ((statusKey!=null) && (statusKey.indexOf("404") != -1))
        	isNotFoundMessage = true;
        else isNotFoundMessage = false;
        
        /** If html-file add ProxyScript */
        if (isHtmlStream && (isNotFoundMessage==false))
        	in = enhanceResponse(in);
        
        /** transmit response (header) */
        sendHeaders(clientOut);
        sendResponse(); /** transmit either all data or, in case of html-data,
        				  * the rest after the proxyScript-part */
        
    }
    
    /** retrieve response headers and store in Hashtable responseHeaders
     *  return remaining input stream for further processing
     */
    private InputStream readResponseHeaders(InputStream in) {
    	
    	byte b 		= 0;
        int len 		= 0;
        int colonIndex	= -1;
        String name, value;
        statusKey = "";
        int returnCode	= 0;
        
        try {
	        if (in == null)
	            in = new BufferedInputStream(serverConnect.getInputStream());
	
	        if (responseHeaders == null) 
	            responseHeaders = new Hashtable();
	            
	        
	        /** Logic to read HTTP response headers */
	        boolean readTooMuch = false;
	        
	  //      System.out.print("");
	
	        for (ByteArrayOutputStream buf = new ByteArrayOutputStream(4097); ;) {
	            if (!readTooMuch) {
	                b = (byte) in.read();
	   //             System.out.print((char) b);
	            }
	            if (b == -1) {
	                break;
	            }
	            readTooMuch = false;
	            if ((b != '\r') && (b != '\n')) {
	                if ((b == ':') && (colonIndex == -1)) {
	                    colonIndex = len;
	                }
	                len++;
	                buf.write(b);
	            } else if (b == '\r') {
	                continue;
	            } else {    // b== '\n'
	                if (len == 0) {
	                    break;
	                }
	                b = (byte) in.read();
	//                System.out.print((char) b);
	                readTooMuch = true;
	
	                /** A space or tab at the begining of a line means the header continues */
	                if ((b == ' ') || (b == '\t')) {
	                    continue;
	                }
	                buf.close();
	                byte[] hdata = buf.toByteArray();
	                buf.reset();
	                if (colonIndex != -1) {
	                    name =
	                            new String(hdata, 0, colonIndex,
	                                    HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING);
	                    value =
	                            new String(hdata, colonIndex + 2, len - 2 - colonIndex,
	                                    HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING);
	                    colonIndex = -1;
	                } else {
	
	                    name =
	                            new String(hdata, 0, len,
	                                    HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING);
	                    value 		= "";
	                    if (name.toLowerCase().indexOf("http")!=-1)
	                    	statusKey 	= name;
	                }
	                
	                String[ ] headerField = {name, value};
	                responseHeaders.put(name.toLowerCase(), headerField);
	                len = 0;
	            }
	        }
	        /** add via-header */
	        java.net.InetAddress i = java.net.InetAddress.getLocalHost();
	        String newVia = "";
	        if (responseHeaders.containsKey(HTTPConstants.HEADER_VIA.toLowerCase())) {
	        	String[] headerFieldVia = (String[])requestHeaders.get(HTTPConstants.HEADER_VIA.toLowerCase());
	        	newVia 	= headerFieldVia[1] + ", ";
	        }
	        newVia = newVia + i.getHostName() + " (UsaProx V1.0)";
	        String[] headerFieldVia = {HTTPConstants.HEADER_VIA, newVia};
	        responseHeaders.put(HTTPConstants.HEADER_VIA.toLowerCase(), headerFieldVia);
        
        } catch (IOException e) {
        	if (serverConnect  != null) {
        		System.err.println("\nAn ERROR occured while reading response headers: "
        					+ "from server " + serverConnect.getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while reading response headers: "
        					+ "\n"
							+ e );
        	}
		}

        return in;
    }
    
    
    /** parsing html-stream and adding ProxyScript after <head>
     **/
    private InputStream enhanceResponse(InputStream in) {
    	
        byte b 					= 0;
        boolean readTooMuch 	= false;
        boolean noHead			= true;
        data 					= new StringBuffer();
        byte[] bdata;
        String dataBefore;
        
        try {
        	
        	/** read in max 4097 bytes and search the head-tag */
	        for (ByteArrayOutputStream buf = new ByteArrayOutputStream(4097); ;) {
	            if (!readTooMuch) {
	                b = (byte) in.read();
	            }
	            if (b == -1) {
	                break;
	            }
	            readTooMuch = false;
	            if (b != '<') {
	                buf.write(b);
	            } else {    // b== '<'
	            	buf.write(b);
	            	b = (byte) in.read();
	            	if ((b != 'h') && (b != 'H')) {
	            		buf.write(b);
	            		continue;
	            	}
	            	/** b== 'h' */
	            	buf.write(b);
	                b = (byte) in.read();
	                if ((b != 'e') && (b != 'E')) {
	                	buf.write(b);
	            		continue;
	            	}
	            	/** b== 'e' */
	            	buf.write(b);
	                b = (byte) in.read();
	                if ((b != 'a') && (b != 'A')) {
	                	buf.write(b);
	            		continue;
	            	}
	            	/** b== 'a' */
	            	buf.write(b);
	            	b = (byte) in.read();
	                if ((b != 'd') && (b != 'D')) {
	                	buf.write(b);
	            		continue;
	            	}
	            	/** b== 'd' */
	            	buf.write(b);
	            	b = (byte) in.read();
	                if (b != '>') {
	                	buf.write(b);
	            		continue;
	            	}
	            	buf.write(b);
	                /** head-tag found */
	                System.out.println("\nUsaProx script tag was added to file " + getKey + "\n"
	                				+ "Monitoring was started at client " + clientConnect.getInetAddress().getHostName()
	                				+ " (" + clientConnect.getInetAddress().getHostAddress() + ")");
	                readTooMuch = true;
	                noHead		= false;
	                buf.close();
	                bdata = buf.toByteArray();
	                buf.reset();
	                
	                dataBefore 		= new String(bdata, HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING);
	                String scriptString		= "<script src='http://lo.lo/proxyscript.js' type='text/javascript'></script>";
	                
	                /** add processed data together with script string to the StringBuffer
	                 *  object that will be transmitted later */
	                data.append(dataBefore).append(scriptString);
	                
	                /** adapt header content-length */
	                if (responseHeaders.containsKey(HTTPConstants.HEADER_CONTENT_LENGTH.toLowerCase())) {
	                	String[] headerFieldLen = (String[])responseHeaders.get(HTTPConstants.HEADER_CONTENT_LENGTH.toLowerCase());
		                int contentLength = Integer.parseInt(headerFieldLen[1]);
		                contentLength += scriptString.getBytes().length;
		                String[] headerFieldLen2 = {HTTPConstants.HEADER_CONTENT_LENGTH,""+contentLength};
		                responseHeaders.put(HTTPConstants.HEADER_CONTENT_LENGTH.toLowerCase(), headerFieldLen2);
	                }
	                break;
	            }
	            
	            /** if no head-tag was found simply add
		         *  processed data to the StringBuffer
		         *  that will be later transmitted */
		        if (noHead) {	
		        	buf.close();
		            bdata = buf.toByteArray();
		            buf.reset();
		            data.append(new String(bdata, HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING));
		        }
        	}
        
        } catch (IOException e) {
        	if (serverConnect  != null) {
        		System.err.println("\nAn ERROR occured while enhancing response: "
        					+ "from server " + serverConnect.getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while enhancing response: "
        					+ "\n"
							+ e );
        	}
		}
        
		/** return the rest of the input stream */ 
        return in;
    }
    
    
    /** forwarding of response headers
	 */
	private void sendHeaders(OutputStream out) {
    	
		clientOut = new DataOutputStream(new BufferedOutputStream(clientOut));
		StringBuffer header = new StringBuffer();
		
		for (java.util.Iterator e = responseHeaders.entrySet().iterator(); e.hasNext();) {
	
	        java.util.Map.Entry me = (java.util.Map.Entry) e.next();
	        Object keyObj = me.getKey();
	        if (null == keyObj) continue;
	        String key = keyObj.toString().trim();
	        
	        /** Set status line to the top of the response message */
	        if (key.startsWith("http")) {
	        	String[] headerField = (String[])me.getValue();
	        	header.insert(0,headerField[0] + "\r\n");
	        } else {
	        	String[] headerField = (String[])me.getValue();
	        	header.append(headerField[0]).append(": ").append(headerField[1]).append("\r\n");
	        }
	    }
		if (debug) System.err.println(header);
		header.append("\r\n");
		
		try {
			clientOut.write(header.toString().getBytes(HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING));
			clientOut.flush();
			
			/** in case HTML is transmitted
			 *  create "serverdata" log entry with index already defined by serverRequest
			 *  as suffix for stored server data */
			if((isHtmlStream && (isNotFoundMessage==false)) && httpTraffic.isLogHttpTraffic()) {
				
				/** httpTraffic log entry */
				logData = dateStamp() + " " + get + " serverdata " + httpTrafficIndex;
				logFile.log(null, logData, clientConnect);
				
				/** store headers to indexed txt-file */
				httpTraffic.store(httpTrafficIndex,"\r\n\r\n##### Response #####\r\n\r\n"+header.toString());
			}
			
		} catch (IOException e) {
			if (clientConnect  != null) {
        		System.err.println("\nAn ERROR occured while sending response headers: "
        					+ "to client " + clientConnect.getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while sending response headers: "
        					+ "\n"
							+ e );
        	}
		}	
    }
    
    
    /** simple forwarding of response data
	 *	used if no html-data is being processed
	 */
	private void sendResponse() {
    	
    	byte response[] = new byte[4096]; /** data buffer */  
        int responseLength;   /** length of response as number of bytes */
        
        try {
        	
        	/** transmit enhanced response data */
	        if (isHtmlStream && (isNotFoundMessage==false)) {
	        	clientOut.write(data.toString().getBytes(HTTPConstants.HEADER_DEFAULT_CHAR_ENCODING));
	        	clientOut.flush();
	        	
				/** store data to indexed txt-file */
				if(httpTraffic.isLogHttpTraffic()) httpTraffic.store(httpTrafficIndex, data.toString());	
	        }
        	
        	while ((responseLength = serverIn.read(response)) != -1) {
				if (responseLength > 0) {
			/**		for(int gg=0; gg<response.length; gg++)
						System.out.print((char)response[gg]);  */
		            clientOut.write (response,0,responseLength);
		            clientOut.flush(); 
		            
		            /** in case HTML is transmitted
			 		  * store data to respective serverdata txt-file */
		            if((isHtmlStream && (isNotFoundMessage==false)) && httpTraffic.isLogHttpTraffic()) {
						/** store data to indexed txt-file */
						httpTraffic.store(httpTrafficIndex, response, 0, responseLength);
					} 
		        }
			}
		} catch (IOException e) {
			if (clientConnect  != null) {
        		System.err.println("\nAn ERROR occured while sending response: "
        					+ "to client " + clientConnect.getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while sending response: "
        					+ "\n"
							+ e );
        	}
		}		
    }
    
    /** Returns String of the form "2004-12-31,23:59:59" */
	public String dateStamp() {
		Calendar now = Calendar.getInstance();
		return now.get(Calendar.YEAR) + "-" + (now.get(Calendar.MONTH)+1) + "-"
		  + now.get(Calendar.DAY_OF_MONTH) + "," + (now.get(Calendar.HOUR_OF_DAY)) + ":" + now.get(Calendar.MINUTE)
		  + ":" + now.get(Calendar.SECOND);
	}
	
	/** Returns the serverData.*/
	public HTTPTraffic getHttpTraffic( ) {
		return httpTraffic;
	}

	/** Returns the serverDataIndex. */
	public int getHttpTrafficIndex() {
		return httpTrafficIndex;
	}

	/** Sets the serverDataIndex. */
	public void setHttpTrafficIndex(int httpTrafficIndex) {
		this.httpTrafficIndex = httpTrafficIndex;
	}

}