/** 
 *  UsaProxy - HTTP proxy for tracking and logging of user interactions on websites
    Copyright (C) 2006  Monika Wnuk - Media Informatics Group at the University of Munich

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

 * 
 * UsaProxy main class
 */

import java.net.*;
import java.io.*; 

public class UsaProxy {
	
	/** UsaProxy port */
    private int 				port;
    private InetAddress 		remoteProxy;
    private int 				remotePort;
    
    final static boolean 		debug = false;
    
    /** Constructor: create UsaProxy with port */
    public UsaProxy (int port) { 

		this.port = port;
	
		/** start UsaProxy-Server */
		proxyStart();
		
    }
    
    /** Constructor: create UsaProxy with port, remote proxy, and remote proxy port
     *  In case UsaProxy is running on a computer within a network, e.g. in an IFX LAN */
	public UsaProxy(int port, InetAddress remoteProxy, int remotePort) {
		
		this.port = port;
		this.remoteProxy = remoteProxy;
		this.remotePort = remotePort;
		
		/** start UsaProxy-Server */
		proxyStart();
		
	}

	/** start UsaProxy-Server */
    private void proxyStart() {

		/** Display message */
		System.out.println("Trying to start UsaProxy at port " + port);
		
		LogFile logFile 		= new LogFile();
		ProxyScript proxyScript = new ProxyScript();
		HTTPTraffic serverData	= new HTTPTraffic();
	
		try {
	
		    /** Create server socket */
		    ServerSocket server = new ServerSocket(port); 
		    
		    /** Display start message */
			System.out.println("UsaProxy started at port " + port);
			System.out.println("UsaProxy ready for accepting incoming connections");
	
		    /** ProxyConnect endless loop */
		    while (true) { 
	
				/** waiting for client */
				Socket clientConnect = server.accept(); 
		
				/** Display message: new client with his address  */
				if (debug) System.out.println("\nNew client connected:\n" + clientConnect.getInetAddress().getHostAddress());
		
				if (remoteProxy != null) {	
					/** new ProxyRequest that will handle client requests */
					new ClientRequest (clientConnect, logFile, proxyScript, serverData, remoteProxy, remotePort);
				}
				else {
					/** new ProxyRequest that will handle client requests */
					new ClientRequest (clientConnect, logFile, proxyScript, serverData); 	
				}
		    } 

		/** Handle IO exception */
		} catch (IOException e) {
        	System.err.println("\nAn ERROR occured while starting UsaProxy:\n"
							+ e );
			System.exit(1);
		}
    }


	public static void main(String[] args) {
		
		try {
			
			/** if remote proxy is specified 
			 *  In case UsaProxy is running on a computer within a network, e.g. in an IFX LAN */
			if (args.length == 3) {	
				InetAddress remoteProxy = null;
	            try {	
	            	remoteProxy = InetAddress.getByName(args[1]);
	            }
	            catch(UnknownHostException e) {
	            	System.err.println("\nAn ERROR occured while connecting to remote proxy:\n"
							+ e );
	                System.err.println("Usage: java UsaProxy <port> [<remote proxy ip-address>] [<remote proxy port>]");
	                System.exit(1);
	            }
	            try {
		            if(Integer.parseInt(args[2]) >= 0 && Integer.parseInt(args[2]) <= 65535) {
			            /** Create UsaProxy with command line argument 1 as port number,
				    	 * remoteProxyAddress as remote proxy, argument 3 as remote proxy port */
						new UsaProxy(Integer.parseInt(args[0]), remoteProxy, Integer.parseInt(args[2]));
		            }
		            else {
		            	System.err.println("\nAn ERROR occured while connecting to remote proxy!\n"
		            					+ "Usage: java UsaProxy <port> [<remote proxy ip-address>] [<remote proxy port>]");
		                System.exit(1);
		            }
	            }
	            catch (NumberFormatException e) {
	            	System.err.println("\nAn ERROR occured while binding UsaProxy to port "
	            					+ Integer.parseInt(args[0]) + " and remote port " + Integer.parseInt(args[2]) + ":\n"
									+ e );
	                System.exit(1);
	            }
		    }
			else if (args.length == 1) {
				try {
					/** Create UsaProxy with command line argument as port number */
					new UsaProxy(Integer.parseInt(args[0]));
				}
				catch (NumberFormatException e) {
					System.err.println("\nAn ERROR occured while binding UsaProxy to port "
	            					+ Integer.parseInt(args[0]) + ":\n"
									+ e );
	            	System.err.println("Usage: java UsaProxy <port>");
	                System.exit(1);
	            }
		    }
			else {
				/** Default port */
				new UsaProxy(8000);
		    }
		}
		catch (Exception e) {
		    System.err.println("\nAn ERROR occured while starting UsaProxy:\n"
									+ e );
		    System.exit(1);
		}
	}
}
